#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Map tile downloading code unit tests.
"""

from functools import wraps

from geotiler.tilenet import TileDownloader

import unittest
from unittest import mock


class TileDownloaderTestCase(unittest.TestCase):
    """
    Map tile downloader base class unit tests.
    """
    def test_setting_cache(self):
        """
        Test tile downloader cache change

        Verify that TileDownloader.fetch_image default cache is removed on
        cache change.
        """
        downloader = TileDownloader()
        fetch_image = downloader.fetch_image
        fetch_image_orig = downloader.fetch_image.__wrapped__

        assert fetch_image != fetch_image_orig

        def cache(f):
            @wraps(f)
            def f1(): return None
            return f1

        downloader.set_cache(cache)
        self.assertEqual(fetch_image_orig, downloader.fetch_image.__wrapped__)


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
GeoTiler example to use Redis based cache for map tiles.
"""

import redis

from geotiler.tilenet import DEFAULT_TILE_DOWNLOADER
from geotiler.cache.redis import RedisCache

import geotiler

import logging
logging.basicConfig(level=logging.DEBUG)

downloader = DEFAULT_TILE_DOWNLOADER

# create the cache connecting to local Redis server
client = redis.Redis('localhost')
cache = RedisCache(client, downloader)

# override cache of default downloader
downloader.set_cache(cache)

bbox = 11.78560, 46.48083, 11.79067, 46.48283
mm = geotiler.Map(extent=bbox, zoom=18)

# render the map for the first time...
img = geotiler.render_map(mm)

# ... and second time to demonstrate use of the cache
img = geotiler.render_map(mm)

# show some recent keys
print('recent cache keys {}'.format(client.keys()[:10]))

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Redis cache unit tests.
"""

from geotiler.cache.redis import RedisCache

import unittest
from unittest import mock

class RedisCacheTestCase(unittest.TestCase):
    """
    Redis cache unit tests.
    """
    def test_wrapper(self):
        """
        Test Redis cache wrapper
        """ 
        client = mock.MagicMock()
        downloader = mock.MagicMock()
        f = lambda host, path, query: True

        cache = RedisCache(client, downloader)
        fc = cache(f)
        self.assertEqual(f, fc.__wrapped__)


    def test_updating_cache(self):
        """
        Test Redis cache update

        Check that valid paramaters are passed to underlying function and
        that cache got updated
        """ 
        client = mock.MagicMock()

        downloader = mock.MagicMock()
        data = mock.MagicMock()
        downloader.f.return_value = data

        cache = RedisCache(client, downloader)
        fc = cache(downloader.f) # function f with cachinig capability

        client.exists.return_value = False
        value = fc('host', 'path', 'query')
        self.assertEqual(data, value)
        downloader.f.assert_called_once_with(
            downloader, 'host', 'path', 'query'
        )
        
        client.setex.assert_called_once_with(
            ('host', 'path', 'query'),
            data,
            cache.timeout
        )


    def test_cache_use(self):
        """
        Test Redis cache use

        Verify that value is fetched from Redis cache on cache hit
        """ 
        client = mock.MagicMock()
        data = mock.MagicMock() # data returned from cache

        downloader = mock.MagicMock()

        cache = RedisCache(client, downloader)
        fc = cache(downloader.f) # function f with cachinig capability

        client.exists.return_value = True # cache hit
        client.get.return_value = data    # return data from cache

        value = fc('host', 'path', 'query')
        self.assertEqual(data, value)
        self.assertFalse(downloader.f.called)
        client.get.assert_called_once_with(('host', 'path', 'query'))


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
GeoTiler map tiles downloading functionality.

To download map tiles and render them as a single image use
:py:func:`render_tiles` function.

The :py:class:`TileThreadDownloader` implements default tiles downloading
strategy. New strategy can be implemented by deriving from
:py:class:`TileDownloader` abstract class.
"""

from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
import http.client
import io
import sys
import traceback
import urllib.parse
import logging

import PIL.Image
import PIL.ImageDraw

logger = logging.getLogger(__name__)

HEADERS = {
    'User-Agent': 'GeoTiler/0.3.0',
}

MAX_ATTEMPTS = 5 # how many times to retry a failing tile


class TileRequest(object):
    """
    Tile request information.

    :var provider: Map tile provider.
    :var coord: Map tile coordinate.
    :var zoom: Map zoom level.
    :var offset: Tile position on map image.
    :var images: List of downloaded tile images.
    :var done: Downloaded if true.
    """
    def __init__(self, provider, coord, zoom, offset):
        self.provider = provider
        self.coord = coord
        self.zoom = zoom
        self.offset = offset
        self.images = []
        self.done = False



class TileDownloader(object):
    """
    Tile downloader abstract class.
    """
    def fetch_tile(self, tile):
        """
        Download tile images.

        The method returns list of images, each being instance of PIL image.

        :param tile: Tile to download.
        """
        if tile.done:
            # don't bother?
            return

        urls = tile.provider.get_tile_urls(tile.coord, tile.zoom)

        if __debug__:
            logger.debug('Requesting {}'.format(urls))

        # this is the time-consuming part
        try:
            images = []

            for (scheme, host, path, params, query, fragment) in map(urllib.parse.urlparse, urls):
                if scheme in ('file', ''):
                    data = open(path)
                else:
                    assert scheme == 'http'
                    data = self.fetch_image(host, path, query)
                    data = io.BytesIO(data)

                img = PIL.Image.open(data).convert('RGBA')
                images.append(img)
                tile.done = True

            if __debug__:
                logger.debug('received {}'.format(urls))

        except Exception as ex:
            # on error log the information and set downloaded images to
            # null; the renderer will handle this
            msg = 'failed to download tile at {}'.format(urls)
            logger.warning(msg)
            msg = ''.join(traceback.format_exception(*sys.exc_info()))
            logger.debug(msg)

            images = [None for url in urls]

        tile.images = images


    @lru_cache()
    def fetch_image(self, host, path, query):
        """
        Fetch an image using HTTP connection.

        The method returns instance of PIL image.

        :param host: HTTP server.
        :param path: HTTP path.
        :param query: HTTP query.
        """
        data = None

        conn = http.client.HTTPConnection(host)
        conn.request('GET', path + ('?' + query).rstrip('?'), headers=HEADERS)
        response = conn.getresponse()
        status = str(response.status)

        if status.startswith('2'):
            data = response.read()

        return data


    def set_cache(self, cache):
        """
        Override existing map tile downloader cache.

        The cache decorator function takes
        :py:func:`TileDownloader.fetch_image` function as an argument and
        creates a wrapper, which caches the results of the
        :py:func:`TileDownloader.fetch_image` function.

        :param cache: Cache decorator function.

        .. seealso:: :py:class:`geotiler.cache.redis.RedisCache`
        """
        self.fetch_image = cache(self.fetch_image.__wrapped__)


    def fetch(self, tiles):
        """
        Execute all tile requests.

        :param tiles: List of tile requests.
        """
        raise NotImplementedError()



class TileThreadDownloader(TileDownloader):
    """
    Tile downloader based on thead pool executor.
    """
    def fetch(self, tiles):
        """
        Execute all tile requests.

        :param tiles: List of tile requests.
        """
        pool = ThreadPoolExecutor(max_workers=32)
        pool.map(self.fetch_tile, tiles, timeout=5)
        pool.shutdown()



def render_tiles(provider, zoom, size, tiles, downloader=None):
    """
    Download map tiles and render them as an image.

    The default tiles downloader is :py:class:`TileThreadDownloader`.

    :param provider: Map tiles provider.
    :param zoom: Map zoom level.
    :param size: Map image size.
    :param tiles: List of tile coordinates and their map image positions.
    :param downloader: Map tiles downloader.
    """
    if downloader is None:
        downloader = DEFAULT_TILE_DOWNLOADER

    tile_req = [TileRequest(provider, t[0], zoom, t[1]) for t in tiles]
    requests = tile_req[:]
    for k in range(MAX_ATTEMPTS):
        downloader.fetch(requests)
        requests = [t for t in requests if not t.done]
        if not requests:
            break

    image = PIL.Image.new('RGB', size)

    for tile in tile_req:
        try:
            for img in tile.images:
                # no image, so there was error when downloading a tile
                # draw error message in such case
                if not img:
                    w, h = provider.tile_width, provider.tile_height
                    img = PIL.Image.new('RGBA', (w, h))
                    draw = PIL.ImageDraw.Draw(img)
                    msg = 'Map tile download error.'
                    tw, th = draw.textsize(msg)
                    draw.text(((w - tw) // 2, (h - th) // 2), msg, 'red')

                image.paste(img, tile.offset, img)
        except Exception as ex:
            logger.warning('tile rendering error')
            logger.exception(ex)

    return image

DEFAULT_TILE_DOWNLOADER = TileThreadDownloader()

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Cache for GeoTiler based on Redis.
"""

from functools import wraps
import logging

logger = logging.getLogger(__name__)

class RedisCache(object):
    """
    Cache for GeoTiler based on Redis.

    :var client: Redis client object.
    :var downloader: Map tile downloader instance.
    :var timeout: Expiry time for the cached data.
    """

    def __init__(self, client, downloader, timeout=3600 * 24 * 7):
        """
        Create Redis based cache.

        :param client: Redis client object.
        :param downloader: Map tile downloader instance.
        :param timeout: Expiry time for the cached data.
        """
        self.client = client
        self.downloader = downloader
        self.timeout = timeout


    def __call__(self, f):
        """
        Wrap tile downloader's function `f` to cache the function results.

        :param f: Map tile downloader function, which results shall be
            cached.
        """
        @wraps(f)
        def wrapper(host, path, query):
            key = (host, path, query)
            if self.client.exists(key):
                if __debug__:
                    logger.debug('cache hit, key {}'.format(key))
                data = self.client.get(key)
                return data
            else:
                data = f(self.downloader, host, path, query)
                self.client.setex(key, data, self.timeout)
                if __debug__:
                    logger.debug('data stored in cache, key {}'.format(key))
                return data
        return wrapper


# vim: sw=4:et:ai
